# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
API test driver mixin.
"""


from __future__ import absolute_import

from . import mixin_test_callables



class TestWiresAPIMixin(mixin_test_callables.TestCallablesMixin):

    """
    Drives Wires API tests requiring mixed class to:
    - Have a Wiring instance at self.w.
    - Allow wiring via self.wire.
    - Allow unwiring via self.unwire.
    """

    def test_unwired_call_does_not_fail(self):
        """
        Calling an unwired call works. Does nothing, but works.
        """
        self.w.unwired_call()


    def test_wiring_non_callable_raises_value_error(self):
        """
        Wiring a call to a non-callable raises ValueError.
        The exception argument (message):
        - Starts with "argument not callable: ".
        - Contains repr(argument).
        """
        for non_callable in (None, True, 42, 2.3, (), [], {}, set()):

            with self.assertRaises(ValueError) as cm:
                self.w.this.wire(non_callable)

            exception_args = cm.exception.args
            self.assertEqual(len(exception_args), 1)

            msg = exception_args[0]
            self.assertTrue(
                msg.startswith('argument not callable: '),
                'wrong exception message: %r' % (msg,),
            )
            self.assertIn(
                repr(non_callable),
                msg,
                'missing argument repr in message: %r' % (msg,),
            )


    def test_unwiring_non_callable_raises_value_error(self):
        """
        Unwiring a call from a non-callable raises ValueError.
        The exception argument (message):
        - Starts with "argument not callable: ".
        - Contains repr(argument).
        """
        for non_callable in (None, True, 42, 2.3, (), [], {}, set()):

            with self.assertRaises(ValueError) as cm:
                self.w.this.unwire(non_callable)

            exception_args = cm.exception.args
            self.assertEqual(len(exception_args), 1)

            msg = exception_args[0]
            self.assertTrue(
                msg.startswith('argument not callable: '),
                'wrong exception message: %r' % (msg,),
            )
            self.assertIn(
                repr(non_callable),
                msg,
                'missing argument repr in message: %r' % (msg,),
            )


    def test_wiring_callable_works(self):
        """
        Wiring a callable works.
        """
        self.w.this.wire(self.returns_42)
        self.addCleanup(self.w.this.unwire, self.returns_42)


    def test_wiring_unwiring_works(self):
        """
        Wiring and then unwiring same callable works.
        """
        self.w.this.wire(self.returns_42)
        self.w.this.unwire(self.returns_42)


    def test_unwiring_unknown_callable_raises_value_error(self):
        """
        Unwiring an unknown callable raises a ValueError.
        The exception argument (message):
        - Starts with "unknown function ".
        """
        with self.assertRaises(ValueError) as cm:
            self.w.this.unwire(self.returns_42)

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)

        msg = exception_args[0]
        self.assertTrue(
            msg.startswith('unknown function '),
            'wrong exception message: %r' % (msg,),
        )


    def test_dynamic_name_wire_unwire_works(self):
        """
        Wiring/unwiring via indexing works.
        """
        name = 'name'
        self.w[name].wire(self.returns_42)
        self.w[name].unwire(self.returns_42)


    def test_create_and_delete_callable_works(self):
        """
        Creating and deleting a callable works.
        """
        self.w.this.wire(self.returns_42)
        del self.w.this


    def test_callables_have_names(self):
        """
        Created callables have a __name__ attribute matching their name.
        """
        self.w.this.wire(self.returns_42)
        self.addCleanup(self.w.this.unwire, self.returns_42)

        self.assertEqual(self.w.this.__name__, 'this')


    def test_callables_wiring_attribute(self):
        """
        Created callables have a `.wirings` attribute that is a list of
        (<wired-callable>, <wired-args-tuple>, <wired-kwargs-dict>).
        """
        self.w.this.wire(self.returns_42)
        self.addCleanup(self.w.this.unwire, self.returns_42)

        wirings = self.w.this.wirings

        self.assertEqual(len(wirings), 1)

        wired_callable, wired_args, wired_kwargs = wirings[0]
        self.assertIs(wired_callable, self.returns_42)
        self.assertIsInstance(wired_args, tuple)
        self.assertEqual(wired_args, ())
        self.assertIsInstance(wired_kwargs, dict)
        self.assertEqual(wired_kwargs, {})


    def test_iteration_works(self):
        """
        Iterating over a Wiring instance produces its callables.
        """
        self.w.callable1.wire(self.returns_42)
        self.w.callable2.wire(self.returns_none)
        self.addCleanup(self.w.callable1.unwire, self.returns_42)
        self.addCleanup(self.w.callable2.unwire, self.returns_none)

        created_callables = set((self.w.callable1, self.w.callable2))
        obtained_callables = set(self.w)

        self.assertEqual(created_callables, obtained_callables)


    def test_wiring_len(self):
        """
        No callables means len is 0.
        """
        self.assertEqual(len(self.w), 0)


    def test_wire_wiring_len(self):
        """
        Rich wiring and callable length tests.
        """
        self.w.callable1.wire(self.returns_42)
        self.addCleanup(self.w.callable1.unwire, self.returns_42)

        self.assertEqual(len(self.w), 1)
        self.assertEqual(len(self.w.callable1), 1)

        self.w.callable2.wire(self.returns_42)
        self.w.callable2.wire(self.returns_none)
        self.addCleanup(self.w.callable2.unwire, self.returns_42)
        self.addCleanup(self.w.callable2.unwire, self.returns_none)

        self.assertEqual(len(self.w), 2)
        self.assertEqual(len(self.w.callable1), 1)
        self.assertEqual(len(self.w.callable2), 2)

        del self.w.callable1
        self.assertEqual(len(self.w), 1)
        self.assertEqual(len(self.w.callable2), 2)


    def test_wiring_repr(self):
        """
        Looks like an instantiation call, which should work.
        """
        repr_str = repr(self.w)
        self.assertTrue(repr_str.startswith('Wiring('), 'bad repr start')
        self.assertTrue(repr_str.endswith(')'), 'bad repr end')

        from wires import Wiring
        eval(repr_str)


    def test_callable_repr(self):
        """
        Between angle brackets, contains callable name repr.
        """
        self.w.callable_name.wire(self.returns_42)
        self.addCleanup(self.w.callable_name.unwire, self.returns_42)

        repr_str = repr(self.w.callable_name)
        self.assertTrue(repr_str.startswith('<WiringCallable '), 'bad repr start')
        self.assertTrue(repr_str.endswith('>'), 'bad repr end')
        self.assertTrue(repr('callable_name') in repr_str, 'no name in repr')


    def test_del_unknown_callable_attr_raises_attribute_error(self):
        """
        Deleting an unknown Callable attribute raises AttributeError.
        """
        with self.assertRaises(AttributeError) as cm:
            del self.w.this.no_such_attribute

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'no_such_attribute')


    def test_setgetdel_unknown_callable_attribute_works(self):
        """
        Can set, get and del random Callable attributes: as expected.
        """
        with self.assertRaises(AttributeError):
            _ = self.w.this.no_such_attribute

        self.w.this.no_such_attribute = 2
        self.assertEqual(self.w.this.no_such_attribute, 2)
        del self.w.this.no_such_attribute

        with self.assertRaises(AttributeError):
            _ = self.w.this.no_such_attribute


    def test_callable_setattr_delattr(self):
        """
        Deleting a Callable attribute reverts its value to its instance's value.
        """
        defaults = {
            'min_wirings': (None, 1),
            'max_wirings': (None, 1),
            'returns': (False, True),
            'ignore_failures': (True, False),
        }
        for attr_name, (default_value, test_value) in defaults.items():
            value = getattr(self.w.this, attr_name)
            self.assertEqual(
                value,
                default_value,
                'starting %s should be %r' % (attr_name, default_value),
            )

            setattr(self.w.this, attr_name, test_value)
            value = getattr(self.w.this, attr_name)
            self.assertEqual(value, test_value)

            delattr(self.w.this, attr_name)
            value = getattr(self.w.this, attr_name)
            self.assertEqual(
                value,
                default_value,
                'final %s should be %r' % (attr_name, default_value),
            )


    def test_callable_set(self):
        """
        Exercises the Callable's .set method.
        """
        # Set multiple settings on the `this` callable.
        self.w.this.set(
            min_wirings=1,
            max_wirings=1,
            returns=True,
            ignore_failures=False,
        )

        # Calling it raises an exception: min_wirings=1 but no wirings.
        with self.assertRaises(RuntimeError) as cm:
            _ = self.w.this()
        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'less than min_wirings wired')

        # Wire one callable.
        self.w.this.wire(self.raises_exception)
        self.addCleanup(self.w.this.unwire, self.raises_exception)

        # Calling it raises a different exception: returns + ignore_failures
        with self.assertRaises(RuntimeError) as cm:
            _ = self.w.this()
        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        # (<callee-exception>, None) are the exception's arguments.
        exception, result = exception_args[0]
        self.assertIsInstance(exception, ValueError)
        self.assertEqual(len(exception.args), 1)
        self.assertEqual(exception.args[0], 'test exception value error message')
        self.assertIsNone(result)

        # Need to set min_wirings back to None such that cleanup does not fail.
        self.w.this.set(min_wirings=None)


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Wires instance argument passing tests.
"""


from __future__ import absolute_import

import unittest

from . import mixin_use_new_instance, mixin_test_args



class TestWiresNoArgPassing(mixin_use_new_instance.UseNewInstanceMixin,
                            mixin_test_args.TestWiresNoArgPassingMixin,
                            unittest.TestCase):

    """
    Wires instance no argument call tests.
    """



class TestWiresWireArgPassing(mixin_use_new_instance.UseNewInstanceMixin,
                              mixin_test_args.TestWiresWireArgPassingMixin,
                              unittest.TestCase):

    """
    Wires instance wire-time positional argument passing call tests.
    """



class TestWiresWireKwargPassing(mixin_use_new_instance.UseNewInstanceMixin,
                                mixin_test_args.TestWiresWireKwargPassingMixin,
                                unittest.TestCase):

    """
    Wires instance wire-time named argument passing call tests.
    """



class TestWiresWireArgKwargPassing(mixin_use_new_instance.UseNewInstanceMixin,
                                   mixin_test_args.TestWiresWireArgKwargPassingMixin,
                                   unittest.TestCase):

    """
    Wires instance wire-time positional and named argument passing call tests.
    """



class TestWiresCallArgPassing(mixin_use_new_instance.UseNewInstanceMixin,
                              mixin_test_args.TestWiresCallArgPassingMixin,
                              unittest.TestCase):

    """
    Wires instance call-time positional argument passing call tests.
    """



class TestWiresCallKwargPassing(mixin_use_new_instance.UseNewInstanceMixin,
                                mixin_test_args.TestWiresCallKwargPassingMixin,
                                unittest.TestCase):

    """
    Wires instance call-time named argument passing call tests.
    """



class TestWiresCallArgKwargPassing(mixin_use_new_instance.UseNewInstanceMixin,
                                   mixin_test_args.TestWiresCallArgKwargPassingMixin,
                                   unittest.TestCase):

    """
    Wires instance call-time positional and named argument passing call tests.
    """



class TestWiresFullPassing(mixin_use_new_instance.UseNewInstanceMixin,
                           mixin_test_args.TestWiresFullPassingMixin,
                           unittest.TestCase):

    """
    Wires instance wire-time and call-time positional and named argument passing
    call tests.
    """



class TestWiresDoubleFullPassing(mixin_use_new_instance.UseNewInstanceMixin,
                                 mixin_test_args.TestWiresDoubleFullPassingMixin,
                                 unittest.TestCase):

    """
    Wires instance double wiring wire-time and call-time positional and named
    argument passing call tests.
    """


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Python Wiring Callable.
"""

from __future__ import absolute_import



class WiringCallable(object):

    """
    Callable with a minimal API:

    - Has zero or more wirings: functions/callables wired to it.
    - Each wiring has optional wire-time arguments.
    - Calling it calls all wired functions/callables, passing them the
      combination of call-time and wire-time arguments.
    """

    def __init__(self, wiring, name, wiring_settings):

        self._wiring = wiring
        self._name = name

        # Default settings, from Wiring instance.
        self._wiring_settings = wiring_settings

        # Per callable settings.
        self._callable_settings = {}

        # Call-time settings.
        self._calltime_settings = {}

        # Wired (<callable>, <wire-time-args>, <wire-time-kwargs>) tuples.
        self._wirings = []


    def __repr__(self):

        return '<%s %r>' % (self.__class__.__name__, self._name)


    @property
    def __name__(self):
        """
        The callable name, like regular functions have.
        """
        return self._name


    # Used as a guard for non-set arguments in the `set` method call; `None`
    # would not be appropriate given than `min_wirings` and `max_wirings` take
    # `None` as valid value.

    not_set = object()

    def set(self, min_wirings=not_set, max_wirings=not_set, returns=not_set,
            ignore_failures=not_set, _next_call_only=False):
        """
        Sets one or more per-callable settings.

        The `not_set` defaults are used as a guard to identify non-set arguments.

        The `_next_call_only` argument is considered private and may be removed
        in future releases; it is used internally to override call-time settings.
        """

        if _next_call_only is True:
            target_settings = self._calltime_settings
        else:
            target_settings = self._callable_settings

        # Going with a `**kwargs` like argument would make this code simpler,
        # but the method signature would be more opaque; we prefer explicit
        # even though the code needs to repeat the argument names and work
        # at a somewhat "meta-ish" level.

        local_names = locals()
        arg_names = ('min_wirings', 'max_wirings', 'returns', 'ignore_failures')
        for name in arg_names:
            if local_names[name] is not self.not_set:
                target_settings[name] = local_names[name]


    @property
    def wirings(self):
        """
        List of wired (<callable>, <wire-time-args>, <wire-time-kwargs>) tuples,
        where <wire-time-args> is a tuple and <wire-time-kwargs> is a dict.
        """
        return list(self._wirings)


    def __len__(self):
        """
        Wiring count.
        """
        return len(self._wirings)


    def _effective_setting(self, setting_name):

        # Call-time settings take precedence over per-Callable settings, which
        # take precedence over Wiring settings.

        return self._calltime_settings.get(
            setting_name,
            self._callable_settings.get(
                setting_name,
                self._wiring_settings[setting_name]
            )
        )


    @property
    def min_wirings(self):
        """
        Minimum number of allowed wirings. None means no limit.
        """
        return self._effective_setting('min_wirings')


    @min_wirings.setter
    def min_wirings(self, value):
        """
        Minimum number of allowed wirings. None means no limit. Must be:
        - A positive integer.
        - Less than or equal to max_wirings
        - Less than or equal to the current number of wirings, if any.
        """
        if value is not None:
            wiring_count = len(self._wirings)
            if value <= 0:
                raise ValueError('min_wirings must be positive or None')
            elif self.max_wirings is not None and value > self.max_wirings:
                raise ValueError('min_wirings must be <= max_wirings')
            elif wiring_count and value > wiring_count:
                raise ValueError('too few wirings')

        self._callable_settings['min_wirings'] = value


    @property
    def max_wirings(self):
        """
        Maximum number of allowed wirings. None means no limit.
        """
        return self._effective_setting('max_wirings')


    @max_wirings.setter
    def max_wirings(self, value):
        """
        Maximum number of allowed wirings. None means no limit. Must be:
        - A positive integer.
        - Greater than or equal to min_wirings
        - Greater than or equal to the current number of wirings, if any.
        """
        if value is not None:
            wiring_count = len(self._wirings)
            if value <= 0:
                raise ValueError('max_wirings must be positive or None')
            elif self.min_wirings is not None and value < self.min_wirings:
                raise ValueError('max_wirings must be >= min_wirings')
            elif wiring_count and value < wiring_count:
                raise ValueError('too many wirings')

        self._callable_settings['max_wirings'] = value


    @property
    def returns(self):
        """
        If True, calling returns list of wired (<exception>, <result>) tuples.
        """
        return self._effective_setting('returns')


    @returns.setter
    def returns(self, value):

        self._callable_settings['returns'] = value


    @property
    def ignore_failures(self):
        """
        If False, stops calling wired callables on the first raised exception.
        """
        return self._effective_setting('ignore_failures')


    @ignore_failures.setter
    def ignore_failures(self, value):

        self._callable_settings['ignore_failures'] = value


    def __delattr__(self, name):
        """
        Removes any per-Callable setting, reverting to the Wiring instance's
        default values; raises ValueError if the resulting settings would be
        invalid.
        """
        # save and discard any current per-callable setting.
        try:
            save_value = self._callable_settings.pop(name)
        except KeyError:
            # not a local setting: fallback to super's __delattr__ and get out.
            super(WiringCallable, self).__delattr__(name)
            return

        # `next_value` would be effective after current value discarding; set
        # it, locally, to validate or trigger failures our setters have in place
        # (for example, can't set min_wirings to 3, with a single wiring).
        next_value = self._effective_setting(name)
        try:
            setattr(self, name, next_value)
        except ValueError:
            # `next_value` isn't valid, revert to `saved_value` and re-raise.
            setattr(self, name, save_value)
            raise

        # if we're here, `next_value` is valid: discard its local version.
        del self._callable_settings[name]


    def wire(self, function, *args, **kwargs):

        """
        Wires `function` with `args` and `kwargs` as wire-time arguments.
        """

        if not callable(function):
            raise ValueError('argument not callable: %r' % (function,))

        # self._max_wirings can be None, meaning "no limit": comparison ok
        if len(self._wirings) == self.max_wirings:
            raise RuntimeError('max_wirings limit reached')

        self._wirings.append((function, args, kwargs))


    def unwire(self, function):

        """
        Unwires `function`.
        If `function` is wired multiple times, just unwires the first wiring.
        """

        if not callable(function):
            raise ValueError('argument not callable: %r' % (function,))

        # self.min_wirings can be None, meaning "no limit": comparison ok
        if len(self._wirings) == self.min_wirings:
            raise RuntimeError('min_wirings limit reached')

        tuples_to_remove = [v for v in self._wirings if v[0] == function]
        if not tuples_to_remove:
            raise ValueError('unknown function %r' % (function,))
        self._wirings.remove(tuples_to_remove[0])


    def __call__(self, *args, **kwargs):

        # Calling with wiring count < `min_wirings`, if set, is an error.
        min_wirings = self.min_wirings
        if min_wirings and len(self._wirings) < min_wirings:
            raise RuntimeError('less than min_wirings wired')

        # Get call coupling behaviour for this call from our Wiring, resetting
        # it, to account for correct "default" vs "overridden" behaviour.
        return_or_raise = self.returns
        ignore_failures = self.ignore_failures
        self._calltime_settings.clear()

        # Will contain (<exception>, <result>) per-wiring tuples.
        call_result = []

        for wired_callable, wire_args, wire_kwargs in self._wirings:
            try:
                combined_args = list(wire_args)
                combined_args.extend(args)
                combined_kwargs = dict(wire_kwargs)
                combined_kwargs.update(kwargs)
                wired_result = wired_callable(*combined_args, **combined_kwargs)
                call_result.append((None, wired_result))
            except Exception as wired_exception:
                call_result.append((wired_exception, None))
                if not ignore_failures:
                    if return_or_raise:
                        raise RuntimeError(*call_result)
                    else:
                        break

        return call_result if return_or_raise else None


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Python Wiring Shell.
"""

from __future__ import absolute_import

from . import _callable



class Wiring(object):

    """
    Python Wiring
    """

    # Holds the default, per-callable, `min_wirings` and `max_wirings` as well
    # as the default caller/callee call coupling behaviour defining `returns`
    # and `ignore_failures` settings.
    #
    # Tracks wired callabes in `_callables` and call-time override settings in
    # `_calltime_settings`.

    def __init__(self, min_wirings=None, max_wirings=None, returns=False,
                 ignore_failures=True):

        if min_wirings is not None and min_wirings <= 0:
            raise ValueError('min_wirings must be positive or None')
        if max_wirings is not None and max_wirings <= 0:
            raise ValueError('max_wirings must be positive or None')
        if min_wirings and max_wirings and min_wirings > max_wirings:
            raise ValueError('max_wirings must be >= min_wirings')

        self._settings = {
            # Default wiring limits.
            'min_wirings': min_wirings,
            'max_wirings': max_wirings,

            # Default call-time coupling behaviour.
            'returns': returns,
            'ignore_failures': ignore_failures,
        }

        # Tracks known Callable instances:
        # - Keys are callable names (this instance's dynamic attributes).
        # - Values are WiringCallable objects.
        self._callables = {}

        # Call time override settings.
        self._calltime_settings = {}


    def __repr__(self):

        return '%s(%s)' % (
            self.__class__.__name__,
            ', '.join('%s=%r' % (k, v) for k, v in self._settings.items())
        )


    def __call__(self, returns=None, ignore_failures=None):
        """
        Used for call-time parameter override.
        """
        self._calltime_settings.clear()

        if returns is not None:
            self._calltime_settings['returns'] = returns
        if ignore_failures is not None:
            self._calltime_settings['ignore_failures'] = ignore_failures

        return self


    def __getattr__(self, name):
        """
        Attribute based access to Callables.
        """
        try:
            the_callable = self._callables[name]
        except KeyError:
            new_callable = _callable.WiringCallable(self, name, self._settings)
            self._callables[name] = new_callable
            the_callable = new_callable

        if self._calltime_settings:
            the_callable.set(_next_call_only=True, **self._calltime_settings)
            self._calltime_settings.clear()

        return the_callable


    def __delattr__(self, name):
        """
        Deletes Callable attributes.
        """
        del self._callables[name]


    def __iter__(self):
        """
        Produces associated Callables.
        """
        return iter(self._callables.values())


    def __len__(self):
        """
        Callable count.
        """
        return len(self._callables)


    def __getitem__(self, name):
        """
        Index based access to Callables.
        """
        return self.__getattr__(name)


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Wires instance min/max wirings tests.
"""


from __future__ import absolute_import

import unittest

from wires import Wiring

from . import mixin_test_callables



class TestInstanceMinMaxWirings(mixin_test_callables.TestCallablesMixin,
                                unittest.TestCase):

    """
    min/max wirings tests for Wires instances.
    """

    def test_non_positive_min_raises_value_error(self):
        """
        Passing N <= 0 as `min_wirings` raises a ValueError.
        """
        with self.assertRaises(ValueError) as cm:
            _ = Wiring(min_wirings=-42)

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'min_wirings must be positive or None')


    def test_non_positive_max_raises_value_error(self):
        """
        Passing N <= 0 as `max_wirings` raises a ValueError.
        """
        with self.assertRaises(ValueError) as cm:
            _ = Wiring(max_wirings=-42)

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'max_wirings must be positive or None')


    def test_min_wirings_greater_than_max_wirings_raises_value_error(self):
        """
        Passing `min_wirings` > `max_wirings` raises a ValueError.
        """
        with self.assertRaises(ValueError) as cm:
            _ = Wiring(min_wirings=42, max_wirings=24)

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'max_wirings must be >= min_wirings')


    def test_min_none_max_none_wire_unwire(self):
        """
        Wiring a single callable works.
        """
        w = Wiring(min_wirings=None, max_wirings=None)

        w.this.wire(self.returns_42)
        w.this.unwire(self.returns_42)


    def test_min_1_wire_unwire_raises_runtime_error(self):
        """
        Wiring then unwiring with min_wirings=1 raises a RuntimeError.
        """
        w = Wiring(min_wirings=1)

        w.this.wire(self.returns_42)
        with self.assertRaises(RuntimeError) as cm:
            w.this.unwire(self.returns_42)

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'min_wirings limit reached')


    def test_max_1_wire_wire_raises_runtime_error(self):
        """
        Wiring two callables with max_wirings=1 raises a RuntimeError.
        """
        w = Wiring(max_wirings=1)

        w.this.wire(self.returns_42)
        with self.assertRaises(RuntimeError) as cm:
            w.this.wire(self.returns_42)

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'max_wirings limit reached')


    def test_min_1_call_raises_runtime_error(self):
        """
        Calling an unwired callable with min_wirings set raises RuntimeError.
        """
        w = Wiring(min_wirings=1)

        with self.assertRaises(RuntimeError) as cm:
            w.this()

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'less than min_wirings wired')


    def test_min_1_wire_call(self):
        """
        Calling a wired callable with min_wirings works.
        """
        w = Wiring(min_wirings=1)

        w.this.wire(self.returns_42)
        result = w.this()

        self.assertIsNone(result)



class TestCallableMinMaxWirings(mixin_test_callables.TestCallablesMixin,
                                unittest.TestCase):

    """
    Per callable min/max wirings tests.
    """

    def setUp(self):

        self.w = Wiring()


    def test_min_wirings_defaults_to_none(self):
        """
        Callable `min_wirings` defaults to None.
        """
        result = self.w.this.min_wirings
        self.assertIsNone(result)


    def test_max_wirings_defaults_to_none(self):
        """
        Callable `max_wirings` defaults to None.
        """
        result = self.w.this.max_wirings
        self.assertIsNone(result)


    def test_non_positive_min_raises_value_error(self):
        """
        Setting callable `min_wirings` to N <= 0 raises a ValueError.
        """
        with self.assertRaises(ValueError) as cm:
            self.w.this.min_wirings = -42

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'min_wirings must be positive or None')


    def test_non_positive_max_raises_value_error(self):
        """
        Setting callable `max_wirings` to N <= 0 raises a ValueError.
        """
        with self.assertRaises(ValueError) as cm:
            self.w.this.max_wirings = -42

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'max_wirings must be positive or None')


    def test_min_wirings_greater_than_max_wirings_raises_value_error(self):
        """
        Setting callable `min_wirings` > `max_wirings` raises a ValueError.
        """
        self.w.this.max_wirings = 24
        with self.assertRaises(ValueError) as cm:
            self.w.this.min_wirings = 42

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'min_wirings must be <= max_wirings')


    def test_max_wirings_less_than_mmin_wirings_raises_value_error(self):
        """
        Setting callable `max_wirings` < `min_wirings` raises a ValueError.
        """
        self.w.this.min_wirings = 42
        with self.assertRaises(ValueError) as cm:
            self.w.this.max_wirings = 24

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'max_wirings must be >= min_wirings')


    def reset_min_max_wirings(self):
        """
        Used in test cleanups.
        """
        self.w.this.min_wirings = None
        self.w.this.max_wirings = None


    def test_min_1_wire_unwire_raises_runtime_error(self):
        """
        Wiring then unwiring with min_wirings=1 raises a RuntimeError.
        """
        self.w.this.min_wirings = 1
        self.w.this.wire(self.returns_42)

        # clean up in the reverse order, otherwise unwiring fails
        self.addCleanup(self.w.this.unwire, self.returns_42)
        self.addCleanup(self.reset_min_max_wirings)

        with self.assertRaises(RuntimeError) as cm:
            self.w.this.unwire(self.returns_42)

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'min_wirings limit reached')


    def test_max_1_wire_wire_raises_runtime_error(self):
        """
        Wiring two callables with max_wirings=1 raises a RuntimeError.
        """
        self.w.this.max_wirings = 1
        self.w.this.wire(self.returns_42)

        # clean up in the reverse order, otherwise unwiring fails
        self.addCleanup(self.w.this.unwire, self.returns_42)
        self.addCleanup(self.reset_min_max_wirings)

        with self.assertRaises(RuntimeError) as cm:
            self.w.this.wire(self.returns_42)

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'max_wirings limit reached')


    def test_min_1_call_raises_runtime_error(self):
        """
        Calling an unwired callable with min_wirings set raises RuntimeError.
        """
        self.w.this.min_wirings = 1
        self.addCleanup(self.reset_min_max_wirings)

        with self.assertRaises(RuntimeError) as cm:
            self.w.this()

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'less than min_wirings wired')


    def test_min_1_wire_call(self):
        """
        Calling a wired callable with min_wirings works.
        """
        self.w.this.min_wirings = 1
        self.w.this.wire(self.returns_42)

        # clean up in the reverse order, otherwise unwiring fails
        self.addCleanup(self.w.this.unwire, self.returns_42)
        self.addCleanup(self.reset_min_max_wirings)

        result = self.w.this()

        self.assertIsNone(result)


    def test_wire_min_2_raises_value_error(self):
        """
        Setting min_wirings < wired callables raises ValueError.
        """
        self.w.this.wire(self.returns_42)
        self.addCleanup(self.w.this.unwire, self.returns_42)

        with self.assertRaises(ValueError) as cm:
            self.w.this.min_wirings = 2

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'too few wirings')


    def test_wire_wire_max_1_raises_value_error(self):
        """
        Setting min_wirings < wired callables raises ValueError.
        """
        self.w.this.wire(self.returns_42)
        self.w.this.wire(self.returns_42)
        self.addCleanup(self.w.this.unwire, self.returns_42)
        self.addCleanup(self.w.this.unwire, self.returns_42)

        with self.assertRaises(ValueError) as cm:
            self.w.this.max_wirings = 1

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'too many wirings')



class TestCombinedMinMaxWirings(mixin_test_callables.TestCallablesMixin,
                                unittest.TestCase):

    """
    Combined wiring and callable min/max wirings tests.
    """

    def test_callable_cant_revert_to_instance_min_wirings(self):
        """
        Deleting callable `min_wirings` fails if the resulting value (the
        Wiring instance's) would leave the callable in an invalid state.
        """
        w = Wiring(min_wirings=2)

        w.this.min_wirings = None

        w.this.wire(self.returns_42)
        self.addCleanup(w.this.unwire, self.returns_42)

        # Resetting per-callable min_wirings should fail: it would fallback
        # to the instance's min_wirings=2, but there are not enough wirings.
        with self.assertRaises(ValueError) as cm:
            del w.this.min_wirings

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'too few wirings')

        # Per-callable min_wirings should be unchanged.
        self.assertIsNone(w.this.min_wirings)


    def test_callable_cant_revert_to_instance_max_wirings(self):
        """
        Deleting callable `max_wirings` fails if the resulting value (the
        Wiring instance's) would leave the callable in an invalid state.
        """
        w = Wiring(max_wirings=1)

        w.this.max_wirings = 2

        w.this.wire(self.returns_42)
        w.this.wire(self.returns_none)
        self.addCleanup(w.this.unwire, self.returns_42)
        self.addCleanup(w.this.unwire, self.returns_none)

        # Resetting per-callable max_wirings should fail: it would fallback
        # to the instance's max_wirings=1, but there are too many wirings.
        with self.assertRaises(ValueError) as cm:
            del w.this.max_wirings

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], 'too many wirings')

        # Per-callable max_wirings should be unchanged.
        self.assertEqual(w.this.max_wirings, 2)


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Shared Wiring instance test mixin.
"""


from __future__ import absolute_import

from wires import w



class UseSharedInstanceMixin(object):

    """
    Wires tests with shared instance mixin.
    """

    @property
    def w(self):
        """
        Per test mixin contract: self.w is a Wiring instance.
        """
        return w


    @staticmethod
    def tearDown():
        """
        Ensures the shared instance has no pending callables.
        """
        existing_callables = set(w)
        for existing_callable in existing_callables:
            delattr(w, existing_callable.__name__)


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Wires instance caller/callee custom coupling tests.
"""


from __future__ import absolute_import

import sys
import unittest

from . import mixin_test_callables, mixin_test_coupling



# About these tests
# -----------------
# This module dynamically generates test classes for all possible instantiation
# time call coupling parameter combinations.
#
# Refer to the documentation in `mixin_test_coupling` for more details.



def _test_class_name(ita):

    # Test class names include `ita` (instantiation-time argument) names and
    # values.

    class_name_part = ''.join(
        '%s%s' % (str(k).capitalize(), str(v).capitalize())
        for k, v in ita.items()
    )
    return 'TestCoupling%s' % (class_name_part,)



def _generate_test_classes():

    # Generates one test class for each combination of possible instantiation
    # time call coupling argument parameters; the empty argument dict is
    # excluded given that it is already accounted for in the explicitly created
    # `TestCouplingMixin` in `mixin_test_coupling`.
    #
    # Uses `mixin_test_coupling.generate_tests` to populate the class with
    # test case sets.

    base_classes = (
        mixin_test_callables.TestCallablesMixin,
        unittest.TestCase,
    )

    this_module = sys.modules[__name__]

    for ita in mixin_test_coupling.CALL_COUPLING_ARG_COMBINATIONS:
        test_class_name = _test_class_name(ita)
        test_class = type(test_class_name, base_classes, {})
        mixin_test_coupling.generate_tests(test_class, ita)
        setattr(this_module, test_class_name, test_class)


# Generate the actual test classes.

_generate_test_classes()


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Python Wires test helpers.
"""


from __future__ import absolute_import



class CallTracker(object):

    """
    Tracks calls to self.
    """

    def __init__(self, returns=None, raises=None):

        # Appended with (args, kwargs) on each call.
        self.call_args = []

        self._returns = returns
        self._raises = raises


    def __call__(self, *args, **kwargs):

        self.call_args.append((args, kwargs))
        if self._raises:
            raise self._raises
        return self._returns


    @property
    def call_count(self):
        """
        Calls to self count.
        """
        return len(self.call_args)


    def reset(self):
        """
        Reset ourselves.
        """
        self.call_args = []



class CallTrackerAssertMixin(object):

    """
    Call tracking assertion methods.
    """

    def assert_called(self, call_tracker, expected_call_arg_list):
        """
        Asserts `call_tracker` was called as many times as there are entries
        in `expected_call_arg_list` and that the arguments match.
        """
        self.assertEqual(
            call_tracker.call_count,
            len(expected_call_arg_list),
            'call count mismatch',
        )
        self.assertEqual(
            call_tracker.call_args,
            expected_call_arg_list,
            'call argument mismatch',
        )


    def assert_single_call_no_args(self, call_tracker):
        """
        Asserts `call_tracker` was called once, with no arguments.
        """
        expected_call_arg_list = [
            ((), {},),
        ]
        self.assert_called(call_tracker, expected_call_arg_list)


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Caller/callee call-time coupling test driver mixin.
"""


from __future__ import absolute_import

from wires import Wiring

from . import mixin_test_callables



# Overview
# --------
# Caller/callee call-time coupling behaviour is defined by two boolean
# parameters: `returns` and `ignore_failures`.
#
# Depending on their values, call-time behaviour is:
#
#                | ignore_failures=False       | ignore_failures=True
# ---------------+-----------------------------+------------------------------
# returns=False  | All wired callables are     | Stops calling wired callables
#                | called.                     | after first exception.
#                | Returns None.               | Returns None
# ---------------+-----------------------------+------------------------------
# returns=True   | All wired callables are     | Stops calling wired callables
#                | called.                     | after first exception.
#                | Returns list of (<e>, <r>)  | On wired callable exceptions
#                | tuples, one per wired       | raises `RuntimeError` with a
#                | callable.                   | list of (<e>, <r>) as arg;
#                |                             | otherwise, returns list of
#                |                             | (<e>, <r>) tuples, one per
#                |                             | wired callable.
# ----------------------------------------------------------------------------
#
#
# Usage
# -----
# The call-time coupling parameters can be used at three levels:
#
# 1. When initializng a `Wiring` instance, via its initializer arguments:
#
#    >>> w = Wiring(returns=<bool>, ignore_failures=<bool>)
#
# 2. As a per-Callable setting:
#
#    >>> w.some_callable.returns = <bool>
#    >>> w.some_callable.ignore_failures = <bool>
#
# 3. At call-time, overriding the instance/callable's default behaviour:
#
#    >>> w(returns=<bool>, ignore_failures=<bool>).some_callable()
#
#
# About the tests
# ---------------
# Testing combinations of initialization-time coupling arguments vs. per-
# callable overriding arguments vs. call-time overriding arguments vs. single-/
# /multi-/failing-/non-failing- wirings is non trivial and quickly leads to a
# combinatorial explosion.
#
# For that reason, call coupling tests are generated/synthesized instead of
# individually hand coded.
#
# The base idea is to run two test cases on every possible scenario:
#
# Case #1 - Wires a single callable that returns 42.
# Case #2 - Wires three callables:
#           - The first returns None.
#           - The second raises a known exception.
#           - The third returns 42.
#
# This module contains two sets of objects:
# - Functions that add combinations of the two-test-set to a given class.
# - A Mixin class used with the bundled shared Wiring instance tests as well
#   as with a non-customized Wiring instance (like most other test Mixins are
#   used).



def _create_test_method(ita, wirings, pca, cta, raises, result, call_counts):

    def call_coupling_test_method(self):
        """
        Generic test method, covering all possible combinations.
        """

        # Replace the test class Wiring instance with a custom initialized one.
        if ita:
            self.w = Wiring(**ita)

        # Set per-callable settings.
        for callable_setting, callable_value in pca.items():
            setattr(self.w.this, callable_setting, callable_value)

        # Wire all wirings in `wire`, assumed to be helpers.CallTrackers.
        for call_tracker in wirings:
            call_tracker.reset()
            self.w.this.wire(call_tracker)
            self.addCleanup(self.w.this.unwire, call_tracker)

        # Do the actual call with call-time argument overriding.
        if raises:
            with self.assertRaises(raises) as cm:
                self.w(**cta).this()
            actual_result = cm.exception.args
        else:
            actual_result = self.w(**cta).this()

        # Assert expected call counts.
        actual_call_counts = [w.call_count for w in wirings]
        self.assertEqual(call_counts, actual_call_counts, 'call count mismatch')

        # Assert expected result.
        self.assertEqual(result, actual_result, 'result mismatch')

    return call_coupling_test_method



def _test_method_name(ita, pca, cta, call):

    # Test method names include instantiation-time argument names and values,
    # as well as per-callable arguments and call-time overriding argument names
    # and values.

    def _dict_to_method_name_part(arg_dict):
        return '_'.join('%s_%s' % (k, v) for k, v in arg_dict.items())

    return 'test_ita_%s_pca_%s_cta_%s_%s' % (
        _dict_to_method_name_part(ita),
        _dict_to_method_name_part(pca),
        _dict_to_method_name_part(cta),
        call,
    )



def _effective_returns(ita, pca, cta):

    # Return the effective value of `returns` based on instantiation-time
    # argument dict `ita`, per-callable argument dict `pca` and call-time
    # override dict `cta`, considering that the default, instantiation time,
    # value is `False`.

    return cta.get('returns', pca.get('returns', ita.get('returns', False)))



def _effective_ignore_failures(ita, pca, cta):

    # Return the effective value of `ignore_failures` based on instantiation-
    # time argument dict `ita`, per-callable argument dict `pca` and call-time
    # override dict `cta`, considering that the default, instantiation time,
    # value is `True`.

    return cta.get(
        'ignore_failures',
        pca.get('ignore_failures', ita.get('ignore_failures', True))
    )



def _create_returns_42_test(test_class, ita, pca, cta):

    # Adds the first test case to `test_class`.
    # - `ita`: instance initialization-time arguments.
    # - `pca`: per-callable arguments
    # - `cta`: call-time argument overrides.

    # The effective `returns` value.
    returns = _effective_returns(ita, pca, cta)

    # Expected `result` depends on `returns`; everything else is constant.
    wirings = [test_class.returns_42]
    raises = None
    result = [(None, 42)] if returns else None
    call_counts = [1]

    # Create and add the test to the class.
    setattr(
        test_class,
        _test_method_name(ita, pca, cta, 'case1_returns_42'),
        _create_test_method(ita, wirings, pca, cta, raises, result, call_counts)
    )



def _create_2in3_raises_test(test_class, ita, pca, cta):

    # Adds the thrid test case to `test_class`.
    # - `ita`: instance initialization-time arguments.
    # - `pca`: per-callable arguments
    # - `cta`: call-time argument overrides.

    # The effective `returns` and `ignore_failures` values.
    returns = _effective_returns(ita, pca, cta)
    ignore_failures = _effective_ignore_failures(ita, pca, cta)

    wirings = [
        test_class.returns_none,
        test_class.raises_exception,
        test_class.returns_42,
    ]
    # Expected `result` and `raises` depend on `returns` and `ignore_failures`.
    if returns:
        if ignore_failures:
            result = [(None, None), (test_class.EXCEPTION, None), (None, 42)]
            raises = None
        else:
            result = ((None, None), (test_class.EXCEPTION, None),)
            raises = RuntimeError
    else:
        result = None
        raises = None
    # Expected `call_counts` depend on `ignore_failures`.
    call_counts = [1, 1, 1] if ignore_failures else [1, 1, 0]

    # Create and add the test to the class.
    setattr(
        test_class,
        _test_method_name(ita, pca, cta, 'case3_2in3_raises_exception'),
        _create_test_method(ita, wirings, pca, cta, raises, result, call_counts)
    )



CALL_COUPLING_ARG_COMBINATIONS = [
    # Purposely excluding the {} entry.
    {'returns': False},
    {'returns': True},
    {'ignore_failures': False},
    {'ignore_failures': True},
    {'returns': False, 'ignore_failures': False},
    {'returns': False, 'ignore_failures': True},
    {'returns': True, 'ignore_failures': False},
    {'returns': True, 'ignore_failures': True},
]



def generate_tests(test_class, ita_filter=None):
    """
    Generates three test case set combinations and adds them to `test_class`.

    Considers all combinations of Wiring instantiation time, per-callable and
    call-time arguments in a combinatoric product, filtering out instantiation
    time combinations based on `ita_filter`; this is used to target test
    generation to varying "starting points".

    Test case set combinations are then generated based on `ita` (instantiation-
    time arguments), `pca` (per-callable arguments) and `cta` (call-time
    argument overrides).
    """
    call_coupling_arg_combinations = [{}]
    call_coupling_arg_combinations.extend(CALL_COUPLING_ARG_COMBINATIONS)

    for ita in call_coupling_arg_combinations:
        if ita_filter is None:
            if ita:
                # no ita_filter: skip test sets with wiring args.
                continue
        else:
            skip = False
            for arg, value in ita_filter.items():
                if ita.get(arg) != value:
                    # skip test sets with differet wiring args
                    skip = True
                    break
            if skip:
                continue
        # create the test methods
        for pca in call_coupling_arg_combinations:
            for cta in call_coupling_arg_combinations:
                _create_returns_42_test(test_class, ita, pca, cta)
                _create_2in3_raises_test(test_class, ita, pca, cta)



class TestCouplingMixin(mixin_test_callables.TestCallablesMixin):

    """
    Drives Wiring call coupling tests.
    """


# Generate tests, considering only default instantiation time arguments.

generate_tests(TestCouplingMixin, ita_filter=None)


# ----------------------------------------------------------------------------
